/****************************************************************************
 * Universidade Federal Fluminense - Instituto de Computao                *
 * Author: Bruno Siqueira                                                   *
 * brunos@midiacom.uff.br                                                   *
 * 2013                                                                     *
 ****************************************************************************/

///
/// \file	LIBR.cc
/// \brief	Implementation of LIBR agent and related classes.
///
/// This is the main file of this software.
///

#include "LIBR.h"


/// Sets packet type.
int hdr_LIBR_pkt::offset_;

/// Sets packet type.
static class LIBRHeaderClass : public PacketHeaderClass {
public:
	LIBRHeaderClass() : PacketHeaderClass("PacketHeader/LIBR",sizeof(hdr_LIBR_pkt)) {
		bind_offset(&hdr_LIBR_pkt::offset_);
	}
} class_rtProtoLIBR_hdr;

/// Establishes Tcl hooks.
static class LIBRClass : public TclClass {
public:
	LIBRClass() : TclClass("Agent/LIBR") {}
	TclObject* create(int argc, const char*const* argv) {
		assert(argc == 5);
		return (new LIBR((nsaddr_t)Address::instance().str2addr(argv[4])));
	}
} class_rtProtoLIBR;


///
/// \brief Creates necessary timers, binds TCL-available variable and do
/// some more initializations.
/// \param na Identifier for the LIBR agent. It will be used as the address
/// of this routing agent.
///
LIBR::LIBR(nsaddr_t na) : Agent(PT_LIBR), state_(IP2ID(na)), pkt_timer_(this) {
	bind("update_interval_", &update_interval_);
	ra_addr_ = na;
 	ra_id_ = IP2ID(na);
 	is_gateway_ = FALSE;
	seq_num_ = 0;
}


///
/// \brief Sends a update message and reschedules the timer.
/// \param e The event which has expired.
///
void
LIBR_UpdateTimer::expire(Event* e) {
	agent_->state().checkup();
	agent_->send_LIBR_pkt();
//	agent_->print_state();
	resched(agent_->update_interval());
}


///
/// \brief Prints out the the simulation trace.
///
void
LIBR::print_state() {
	state().print(logtarget_);
}


///
/// \brief Interface with TCL interpreter.
///
/// From your TCL scripts or shell you can invoke commands on this LIBR
/// routing agent thanks to this function. Currently you can call "start",
/// "print_rtable", "print_state", "set_gateway", "port-dmux",
/// "log-target" and "tracetarget" commands.
///
/// \param argc Number of arguments.
/// \param argv Arguments.
/// \return TCL_OK or TCL_ERROR.
///
int
LIBR::command(int argc, const char*const* argv) {
	if (argc == 2) {
		if (strcasecmp(argv[1], "start") == 0) {
			pkt_timer_.resched(0.0);
			return TCL_OK;
		}
		else if (strcasecmp(argv[1], "print_rtable") == 0) {
			if (logtarget_ != 0) {
				sprintf(logtarget_->pt_->buffer(),
					"P %f _%d_ Routing Table",
					CURRENT_TIME,
					ra_addr());
				logtarget_->pt_->dump();
				state().rtable().print(logtarget_);
			}
			else {
				fprintf(stdout, "%f _%d_ If you want to print this"
					"routing table, you must create a trace "
					"file in your tcl script",
					CURRENT_TIME,
					ra_addr());
			}
			return TCL_OK;
		}
 		else if (strcasecmp(argv[1], "print_state") == 0) {
			if (logtarget_ != 0) {
				sprintf(logtarget_->pt_->buffer(),
					"P %f _%d_ Node state",
					CURRENT_TIME,
					ra_addr());
				logtarget_->pt_->dump();
				state().print(logtarget_);
			}
			else {
				fprintf(stdout, "%f _%d_ If you want to print this"
					"state, you must create a trace "
					"file in your tcl script",
					CURRENT_TIME,
					ra_addr());
			}
			return TCL_OK;
		}
		else if (strcasecmp(argv[1], "set_gateway") == 0) {
			is_gateway() = TRUE;
			state().pri_gw()->id() = ra_id();
			state().pri_gw()->next_hop() = ra_id();
			state().pri_gw()->metric() = 0;
			std::cout << "Gateway ID: " << (unsigned int) state().pri_gw()->id() << std::endl;
			return TCL_OK;
		}
	}
	else if (argc == 3) {
		// Obtains corresponding dmux to carry packets to upper layers
		if (strcmp(argv[1], "port-dmux") == 0) {
			dmux_ = (PortClassifier*)TclObject::lookup(argv[2]);
			if (dmux_ == 0) {
				fprintf(stderr,
					"%s: %s lookup of %s failed\n",
					__FILE__,
					argv[1],
					argv[2]);
				return TCL_ERROR;
			}
			return TCL_OK;
		}
		// Obtains corresponding tracer
		else if (strcmp(argv[1], "log-target") == 0 ||
			strcmp(argv[1], "tracetarget") == 0) {
			logtarget_ = (Trace*)TclObject::lookup(argv[2]);
			if (logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}
	}
	// Pass the command to the base class
	return Agent::command(argc, argv);
}


///
/// \brief	This function is called whenever a packet is received. It identifies
///		the type of the received packet and process it accordingly.
///
/// If it is an LIBR packet then it is processed. In other case, if it is a data packet
/// then it is forwarded.
///
/// \param	p the received packet.
/// \param	h a handler (not used).
///
void
LIBR::recv(Packet* p, Handler* h) {
	struct hdr_cmn* ch 	= HDR_CMN(p);
	struct hdr_ip* ih 	= HDR_IP(p);

	if (ih->saddr() == ra_addr()) {
		// If there exists a loop, must drop the packet
		if (ch->num_forwards() > 0) {
			drop(p, DROP_RTR_ROUTE_LOOP);
			return;
		}
		// else if this is a packet I am originating, must add IP header
		else if (ch->num_forwards() == 0)
			ch->size() += IP_HDR_LEN;
	}

	// If it is a LIBR packet, must process it
	if (ch->ptype() == PT_LIBR)
		recv_LIBR_pkt(p);
	// Otherwise, must forward the packet (unless TTL has reached zero)
	else {
		ih->ttl_--;
		if (ih->ttl_ == 0) {
			drop(p, DROP_RTR_TTL);
			return;
		}
		forward_data(p);

	}
}


///
/// \brief Processes an incoming LIBR packet.
/// \param p received packet.
///
void
LIBR::recv_LIBR_pkt(Packet* p) {
	struct hdr_ip* ih		= HDR_IP(p);
	struct hdr_LIBR_pkt* lh	= HDR_LIBR_PKT(p);

	// All routing messages are sent from and to port RT_PORT,
	// so we check it.
	assert(ih->sport() == RT_PORT);
	assert(ih->dport() == RT_PORT);
	
	state().update(lh);

	if ( state().new_pri_gw() ) {
		send_LIBR_pkt();
		state().new_pri_gw() = FALSE;
	}

	// Release resources
	Packet::free(p);
}


///
/// \brief Creates LIBR packet according to gateways and neighbors states.
///
void
LIBR::send_LIBR_pkt() {
	Packet* 				p 	= allocpkt();
	struct hdr_cmn*			ch 	= HDR_CMN(p);
	struct hdr_ip*			ih 	= HDR_IP(p);
	struct hdr_LIBR_pkt*	lh	= HDR_LIBR_PKT(p);
	LIBR_nb_it 				it;	
	int 					n;

	seq_num()++;

	// LIBR header
	lh->pkt_src() 		= ra_id();
	lh->pkt_seq_num() 	= seq_num();
	if ( is_gateway() )
		state().pri_gw()->seq() = seq_num();
	lh->msg_pri_gw().id()		= state().pri_gw()->id();
	lh->msg_pri_gw().seq()		= state().pri_gw()->seq();
	lh->msg_pri_gw().next_hop()	= state().pri_gw()->next_hop();
	lh->msg_pri_gw().metric()	= state().pri_gw()->metric();
	lh->msg_sec_gw().id()		= state().sec_gw()->id();
	lh->msg_sec_gw().seq()		= state().sec_gw()->seq();
	lh->msg_sec_gw().next_hop()	= state().sec_gw()->next_hop();
	lh->msg_sec_gw().metric()	= state().sec_gw()->metric();

	// Restrict announced neighbors to LIBR_MAX_NB
	it = state().nb().begin();
	for ( n = 0 ; n < LIBR_MAX_NB ; n++  ) {
		if ( it != state().nb().end() ) {

			lh->msg_nb(n).id()		 = state().nb_id(it);

			if ( state().nb(it).local_ack().timeouts() >= LIBR_INACTIVE_NODE ) {
				lh->msg_nb(n).delivery() = 0;
			}
			else {
				lh->msg_nb(n).delivery() = state().nb(it).local_ack().ratio();
			}

			it++;
		}
		else {
			lh->msg_nb(n).id() 			= NULL_ID;
		}
	}

	// Channel header
	ch->ptype()			= PT_LIBR;
	ch->direction()		= hdr_cmn::DOWN;
	ch->size() 			= IP_HDR_LEN + LIBR_HDR_LEN + n*LIBR_NB_LEN;
	ch->error() 		= 0;
	ch->next_hop() 		= IP_BROADCAST;
	ch->addr_type() 	= NS_AF_INET;

	// IP header
	ih->saddr() 		= ra_addr();
	ih->daddr() 		= IP_BROADCAST;
	ih->sport() 		= RT_PORT;
	ih->dport() 		= RT_PORT;
	ih->ttl() 			= IP_DEF_TTL;

	Scheduler::instance().schedule(target_, p, JITTER);
}


///
/// \brief Forwards a data packet to the appropiate next hop indicated by the routing table.
///
/// \param p the packet which must be forwarded.
///
void
LIBR::forward_data(Packet* p) {
	struct hdr_cmn* ch 	= HDR_CMN(p);
	struct hdr_ip* ih 	= HDR_IP(p);
	LIBR_nb_it	   it;

	if (ch->direction() == hdr_cmn::UP && 
		((u_int32_t)ih->daddr() == IP_BROADCAST || ih->daddr() == ra_addr())) {
		// Reverse Path implementation (begin)
		// Check if source address is neighbor

		it = state().nb().find( IP2ID(ih->saddr()) );
		if  (it == state().nb().end() && // Neighbor not found
			(ih->daddr() == ID2IP(state().pri_gw()->id()) ||
			 ih->daddr() == ID2IP(state().sec_gw()->id()))) {
			state().rtable().add_entry( ih->saddr() , (nsaddr_t) ch->prev_hop_ ); // IP = MAC
		}

		dmux_->recv(p, 0);
		return;
	}
	else {
		ch->direction() = hdr_cmn::DOWN;
		ch->addr_type() = NS_AF_INET;
		if ((u_int32_t)ih->daddr() == IP_BROADCAST)
			ch->next_hop() = IP_BROADCAST;
		else {
			nsaddr_t next_hop = state().rtable().lookup(ih->daddr());
			if ((u_int32_t)next_hop == IP_BROADCAST) {
				next_hop = state().rtable().lookup(state().closestID(ih->daddr()));	
				if ((u_int32_t)next_hop == IP_BROADCAST) {
					debug("%f: Agent %d can not forward a packet destined to %d\n",
						CURRENT_TIME,
						ra_addr(),
						ih->daddr());
					drop(p, DROP_RTR_NO_ROUTE);
					return;
				}
				else
					ch->next_hop() = next_hop;
					ch->prev_hop_ = ra_addr();
			}
			else {
				// Reverse Path implementation (begin)
				// Check if source address is neighbor

				it = state().nb().find( IP2ID(ih->saddr()) );
				if  (it == state().nb().end() && // Neighbor not found
					(ih->daddr() == ID2IP(state().pri_gw()->id()) ||
					 ih->daddr() == ID2IP(state().sec_gw()->id()))) {
					state().rtable().add_entry( ih->saddr() , (nsaddr_t) ch->prev_hop_ ); // IP = MAC
				}

				
				ch->next_hop() = next_hop;
				ch->prev_hop_ = ra_addr();
				// Reverse Path implementation (end)
			}
		}
		Scheduler::instance().schedule(target_, p, 0);
	}
}
