/****************************************************************************
 * Universidade Federal Fluminense - Instituto de Computao                *
 * Author: Bruno Siqueira                                                   *
 * brunos@midiacom.uff.br                                                   *
 * 2013                                                                     *
 ****************************************************************************/

///
/// \file	LIBR.h
/// \brief	Header file for LIBR agent and related classes.
///
/// Here are defined the LIBR timer and the class LIBR.
///

#ifndef __LIBR_h__
#define __LIBR_h__

#include <agent.h>
#include <random.h>
#include <address.h>
#include <classifier-port.h>
#include <cmu-trace.h>
#include "LIBR_state.h"

/// Gets current time from the scheduler.
#define CURRENT_TIME Scheduler::instance().clock()
/// Random number between [0-0.5] used to jitter LIBR packet transmission.
#define JITTER (Random::uniform()*0.5)


class LIBR; // forward declaration


/********** Timer **********/

/// Timer for sending a LIBR update message.
class LIBR_UpdateTimer : public TimerHandler {

	LIBR*	agent_;

protected:
 	virtual void expire(Event* e);

public:
	 LIBR_UpdateTimer(LIBR* agent) : TimerHandler() {
		 agent_ = agent;
	 }

};


/********** Agent **********/

///
/// \brief Routing agent which implements LIBR protocol.
///
/// Interacts with TCL interface through command() method. It implements all
/// functionalities related to sending and receiving packets and managing
/// internal state.
///
class LIBR : public Agent {

	friend class	LIBR_UpdateTimer;
	friend class	LIBR_state;
	
	nsaddr_t 	ra_addr_;
	libr_id_t	ra_id_;
	LIBR_state	state_;
	u_int8_t 	seq_num_;
	bool		is_gateway_;
	nsaddr_t 	reverse_path_;
	double 		update_interval_;

protected:

	PortClassifier*		dmux_; 		// For passing packets up to agents.
	Trace*				logtarget_; // For logging.
	LIBR_UpdateTimer	pkt_timer_; // Timer for sending packets.
	
	inline nsaddr_t&	ra_addr()			{ return ra_addr_; }
 	inline libr_id_t&	ra_id()				{ return ra_id_; }
	inline LIBR_state&	state()				{ return state_; }
 	inline u_int8_t&	seq_num()			{ return seq_num_; }
  	inline bool&		is_gateway()		{ return is_gateway_; }	
	inline nsaddr_t&	reverse_path()		{ return reverse_path_; }
	inline double&		update_interval()	{ return update_interval_; }
	
	void forward_data(Packet*);
	void recv_LIBR_pkt(Packet*);
	void send_LIBR_pkt();
	void print_state();
	
public:
	
	LIBR(nsaddr_t);
	int command(int, const char*const*);
	void recv(Packet*, Handler*);

};
	
#endif  /* __LIBR_h__ */
