/****************************************************************************
 * Universidade Federal Fluminense - Instituto de Computao                *
 * Author: Bruno Siqueira                                                   *
 * brunos@midiacom.uff.br                                                   *
 * 2013                                                                     *
 ****************************************************************************/

///
/// \file	LIBR_repositories.cc
/// \brief	Implementation of the Class BitBucket, used for computation of
/// delivery ratio.
///

#include "LIBR_repositories.h"

///
/// \brief Initializes bucket variables.
///
LIBR_BitBucket::LIBR_BitBucket() {

	ratio_          = 0;
	last_seq_       = 0;
	top_			= 0;
	sum_			= 0;
	timeouts_		= 0;
	wait_			= LIBR_PKT_WINDOW - 1;
	is_first_fill_  = TRUE;

}

///
/// \brief Fills the bit bucket (FILO) of LIBR packets received from a neighbor,
/// where bit 1 represents success and bit 0 represents fail.
///
/// The regular number of bits (or sequence numbers) computed is LIBR_PKT_WINDOW.
///
/// \param seq_num Sequence number of the LIBR message received.
///
void
LIBR_BitBucket::fill(u_int8_t seq_num) {

	// Exception: first filling.
	if ( is_first_fill_ ) {
		last_seq_		= seq_num;
		bucket_			= 1;
		sum_			= 1;
		ratio_			= 1;
		timeouts_		= 0 ;
		is_first_fill_	= 0;
	}
	// Check if it is a newer sequence number (exception: seq_num overflow).
	else if ( ( u_int8_t ( seq_num - last_seq_ - 1 ) ) < LIBR_PKT_WINDOW ) {
		top_ += (u_int8_t) ( seq_num - last_seq_ ) ;
		bucket_ |= ( 1 << top_ );
		sum_ ++ ;
		// Exception: number of valid bits different from LIBR_PKT_WINDOW
		if ( wait_ || top_ >= LIBR_PKT_WINDOW )
			ratio_ = (float) sum_ / (top_ + 1.0) ;
		else
			ratio_ = (float) sum_ / LIBR_PKT_WINDOW ;
		last_seq_ = seq_num ;
		timeouts_ = 0 ;
	}
}

///
/// \brief Drops periodicaly the first bit of the bucket.
///
void
LIBR_BitBucket::drop() {

	// Exception: wait for LIBR_PKT_WINDOW before dropping bits.
	if ( wait_ ) {
		ratio_ = (float) sum_ / (top_ + timeouts_ + 1.0) ;
		wait_ --;
	}
	else if ( top_ ) {
		sum_ -= ( bucket_ & 1 );
		bucket_ >>= 1 ;
		top_ -- ;
		// Exception: number of valid bits different from LIBR_PKT_WINDOW
		if ( top_ >= LIBR_PKT_WINDOW )
			ratio_ = (float) sum_ / (top_ + 1.0) ;
		else
			ratio_ = (float) sum_ / LIBR_PKT_WINDOW ;
	}

	// Timeout counter for waiting period.
	timeouts_ ++ ;

}

