/****************************************************************************
 * Universidade Federal Fluminense - Instituto de Computao                *
 * Author: Bruno Siqueira                                                   *
 * brunos@midiacom.uff.br                                                   *
 * 2013                                                                     *
 ****************************************************************************/

///
/// \file	LIBR_repositories.h
/// \brief	Here are defined all data structures and classes needed by
/// a LIBR node.
///

#ifndef __LIBR_repositories_h__
#define __LIBR_repositories_h__

#include "LIBR_pkt.h"
#include <map>


/// Computation of acks and delivery ratio
class LIBR_BitBucket {

	u_int32_t	bucket_;		///< Bucket of bits (acks)
	u_int8_t   	top_;			///< Top of bucket (first position is 0)
	u_int8_t    sum_;			///< Sum of bits (acks)
	u_int8_t	wait_;			///< Waiting period before start to drop
	float		ratio_;			///< Total bits div by sum of bits
	u_int8_t    last_seq_;		///< Last sequence number
	u_int8_t	timeouts_;		///< Num of no fillings in sequence
	bool        is_first_fill_; ///< Check if it is the irst time of filling

public:

	LIBR_BitBucket();					///< Constructor
	void				fill(u_int8_t);	///< Fill with bit according to seq number
	void				drop();			///< Drain bit
	inline float		ratio()			{ return ratio_; }
	inline u_int8_t		timeouts()		{ return timeouts_; }
	u_int8_t			last_seq()		{ return last_seq_; }
	
	u_int8_t   			top()			{ return top_; }
	u_int8_t    		sum()			{ return sum_; }
	u_int8_t			wait()			{ return wait_; }


};


/// Delivery of neighbors
typedef std::map <libr_id_t, float> 			LIBR_nb_ack;

/// Iterator of neighbors' delibery
typedef std::map <libr_id_t, float>::iterator	LIBR_ack_it;


/// A neighbor tuple
struct LIBR_nb_info {

	libr_id_t		next_hop_;		///< Next hop ID
	float			metric_;		///< Metric to neighbor (ETX)
	LIBR_BitBucket	local_ack_;		///< Local delivery statistic
	LIBR_nb_ack		remote_ack_;	///< Neighbor delivery statistic
	
	inline libr_id_t&		next_hop()					{ return next_hop_; }
	inline float&			metric()					{ return metric_; }
	inline LIBR_BitBucket&	local_ack()					{ return local_ack_; }
	inline LIBR_nb_ack&		remote_ack()				{ return remote_ack_; }
	inline float&			remote_ack(LIBR_ack_it i)	{ return i->second; }
	
};

/// State of neighbors
typedef std::map <libr_id_t, LIBR_nb_info> 				LIBR_nb_state;

/// Iterator of neighbors state
typedef std::map <libr_id_t, LIBR_nb_info>::iterator	LIBR_nb_it;

/// A gateway tuple
struct LIBR_gw_state{

	libr_id_t		id_;				///< ID of gateway
	libr_id_t		next_hop_;			///< Next hop ID of route
	float			metric_;			///< Metric to neighbor (ETX)
	libr_id_t		alt_hop_;			///< ID of alternative next hop
	float			alt_metric_;		///< Metric to next hop (ETX)
	u_int8_t		seq_;				///< Sequence number of message
	u_int8_t		timeouts_;			///< Number of timeouts
	u_int8_t		hop_timeouts_;		///< Number of timeouts of next hop

	inline libr_id_t&		id()				{ return id_; }
	inline libr_id_t&		next_hop()			{ return next_hop_; }
	inline float&			metric()			{ return metric_; }
	inline libr_id_t&		alt_hop()			{ return alt_hop_; }
	inline float&			alt_metric()		{ return alt_metric_; }
	inline u_int8_t&		seq()				{ return seq_; }
	inline u_int8_t&		timeouts()			{ return timeouts_; }
	inline u_int8_t&		hop_timeouts()		{ return hop_timeouts_; }

};

#endif  /* __LIBR_repositories_h__ */
