/****************************************************************************
 * Universidade Federal Fluminense - Instituto de Computao                *
 * Author: Bruno Siqueira                                                   *
 * brunos@midiacom.uff.br                                                   *
 * 2013                                                                     *
 ****************************************************************************/

///
/// \file	LIBR_rtable.cc
/// \brief	Implementation of LIBR routing table.
///

#include "LIBR_rtable.h"

///
/// \brief Creates a new empty routing table.
///
LIBR_rtable::LIBR_rtable() { }

///
/// \brief Prints out the content of the routing table to a given trace file.
///
/// Content is represented as a table in which each line is preceeded by a 'P'.
/// First line contains the name of every column (dest, next) and the
/// following ones are the values of each entry.
///
/// \param out the Trace where the routing table must be written into.
///
void
LIBR_rtable::print(Trace* out) {
	sprintf(out->pt_->buffer(), "P\tdest\tnext");
	out->pt_->dump();
	for (rtable_t::iterator it = rt_.begin(); it != rt_.end(); it++) {
		sprintf(out->pt_->buffer(), "P\t%d\t%d",(*it).first,(*it).second);
		out->pt_->dump();
	}
}

///
/// \brief Clears the routing table and frees the memory assigned to each entry.
///
void
LIBR_rtable::clear() {
	rt_.clear();
}

///
/// \brief Deletes the entry whose destination address is given.
/// \param dest	address of the destination node.
///
void
LIBR_rtable::rm_entry(nsaddr_t dest) {
	rt_.erase(dest);
}

///
/// \brief Adds a new entry into the routing table.
///
/// If an entry for the given destination existed, it is overwritten.
///
/// \param dest		address of the destination node.
/// \param next		address of the next hop node.
///
void
LIBR_rtable::add_entry(nsaddr_t dest, nsaddr_t next) {
	rt_[dest] = next;
}

///
/// \brief Looks up an entry for the specified destination address.
/// \param dest	destination address.
/// \return	the routing table entry for that destination address, or NULL
///		if such an entry does not exist
///
nsaddr_t
LIBR_rtable::lookup(nsaddr_t dest) {
	rtable_t::iterator it = rt_.find(dest);
	if (it == rt_.end())
		return (u_int32_t) IP_BROADCAST;
	else
		return (*it).second;
}

///
/// \brief Returns the number of entries in the routing table.
/// \return the number of entries in the routing table.
///
u_int32_t
LIBR_rtable::size() {
	return rt_.size();
}
