/****************************************************************************
 * Universidade Federal Fluminense - Instituto de Computao                *
 * Author: Bruno Siqueira                                                   *
 * brunos@midiacom.uff.br                                                   *
 * 2013                                                                     *
 ****************************************************************************/

///
/// \file	LIBR_state.cc
/// \brief	Implementation of all functions needed for manipulating the internal
///		state of an LIBR node.
///

#include "LIBR_state.h"

///
/// \brief Initializes the node ID and the primary and secondary gateways.
///
LIBR_state::LIBR_state(libr_id_t id) {
	id_ = id;	
	pri_gw_ = new LIBR_gw_state;
	sec_gw_ = new LIBR_gw_state;
	pri_gw_->id_ = sec_gw_->id_ = NULL_ID;
	pri_gw_->next_hop_ = sec_gw_->next_hop_ = NULL_ID;
	pri_gw_->metric_ = sec_gw_->metric_ = INF_METRIC;
	pri_gw_->seq_ = sec_gw_->seq_ = 0;
	pri_gw_->timeouts_ = sec_gw_->timeouts_ = 0;
	new_pri_gw_ = FALSE;
}

///
/// \brief Updates neighbor and gateways states according to a LIBR message
/// received and writes into routing table.
///
/// \param hdr content of a LIBR message received from a neighbor.
///
void
LIBR_state::update(hdr_LIBR_pkt* hdr) {

	LIBR_nb_it	it;
	float		hop_metric;

	it = nb().find( hdr->pkt_src() );
	
	// New neighbor
	if ( it == nb().end() ) {

		nb_[ hdr->pkt_src() ] ;
		it = nb().find( hdr->pkt_src() );

		hop_metric = updateNeighboor(hdr, it) ;
		updateGateways(hdr, hop_metric);
		updateRoutes(it);

	}
	else if ( newSeqNum( hdr->pkt_seq_num() , nb(it).local_ack().last_seq() ) ) {

		hop_metric = updateNeighboor(hdr, it) ;
		updateGateways(hdr, hop_metric);
		updateRoutes(it);
	}
	
}


///
/// \brief Checks and deletes inactive nodes and changes next hop if necessary.
///
void
LIBR_state::checkup() {

	LIBR_gw_state*	temp;
	LIBR_ack_it		ita;
	libr_id_t		off[LIBR_MAX_NB];
	LIBR_nb_it		del[LIBR_MAX_NB];
	int				n = 0;
	int				m = 0;

	// Drops delivery bit of each node and checks inactive nodes.
	for ( LIBR_nb_it it = nb().begin(); it != nb().end(); ++it ) {

		nb(it).local_ack().drop();

		if ( nb(it).local_ack().timeouts() == LIBR_DELETED_NODE ) {
			del[m++] = it;
		}
		else if ( nb(it).local_ack().timeouts() == LIBR_INACTIVE_NODE ) {
			off[n++] = nb_id(it);
			nb(it).metric() = INF_METRIC;
			rtable().rm_entry(ID2IP(nb_id(it)));
		}
	}

	// Deletes inactive nodes
	for ( int i = 0 ; i < m ; i ++ ) {
		nb().erase(del[i]);
	}

	// Changes next hop if it is inactive
	for ( LIBR_nb_it it = nb().begin(); it != nb().end(); ++it ) {

		if ( nb(it).next_hop() != nb_id(it) ) {

			for ( int i = 0 ; i < n ; i ++ ) {

				if ( nb(it).next_hop() == off[i] ) {
					nb(it).next_hop() = nb_id(it);
					ita = nb(it).remote_ack().find(id());
					
					// Computes new metric.
					if (ita != nb(it).remote_ack().end() &&
						nb(it).local_ack().ratio() != 0  &&
						nb(it).remote_ack(ita) != 0 ) {
						nb(it).metric() = float ( 1.0 / ( nb(it).local_ack().ratio() * nb(it).remote_ack(ita) ));
					}
					else {
						nb(it).metric() = INF_METRIC;
					}
					
					// Writes new next hop into routing table.
					rtable().add_entry( ID2IP(nb_id(it)) , ID2IP(nb(it).next_hop()) );
				}
			}
		}
	}

	// Increments primary gateway timeout.
	if ( pri_gw()->id() != id() )
		pri_gw()->timeouts() ++;

	// Removes inactive gateway from routing table or deletes its state.
	if ( pri_gw()->timeouts() == LIBR_DELETED_NODE ) {
		pri_gw()->id() = NULL_ID;
	}
	else if ( pri_gw()->timeouts() == LIBR_INACTIVE_NODE ) {
		pri_gw()->metric() = INF_METRIC;
		rtable().rm_entry(ID2IP(pri_gw()->id()));
	}

	// Increment secondary gateway timeout.
	sec_gw()->timeouts() ++;

	// Removes inactive gateway from routing table or deletes its state.
	if ( sec_gw()->timeouts() == LIBR_DELETED_NODE ) {
		sec_gw()->id() = NULL_ID;
	}
	else if ( sec_gw()->timeouts() == LIBR_INACTIVE_NODE ) {
		sec_gw()->metric() = INF_METRIC;
		rtable().rm_entry(ID2IP(sec_gw()->id()));
	}

	// Reorders primary and secondary gateways according to the best metric.
	if ( sec_gw()->metric() < pri_gw()->metric() ) {
		temp = pri_gw();
		pri_gw() = sec_gw();
		sec_gw() = temp;
	}

}


///
/// \brief Updates neighbor's metric and next hop and returns the best metric.
///
/// \param hdr content of a LIBR message received from a neighbor
/// \param it map iterator with position of related neighbor.
///
/// \return Best metric of related neighbor.
float
LIBR_state::updateNeighboor(hdr_LIBR_pkt* hdr, LIBR_nb_it it) {

	LIBR_nb_it	itn;
	LIBR_ack_it	ita;
	float		alt;
	float		best_metric;
	libr_id_t	best_hop;
	
	// Updates delivery  ratio
	nb(it).local_ack().fill(hdr->pkt_seq_num());
	nb(it).remote_ack().clear();

	best_metric	= INF_METRIC;
	best_hop	= hdr->pkt_src();

	for ( int i = 0 ; i < LIBR_MAX_NB && hdr->msg_nb(i).id() != NULL_ID ; i++ ) {

		nb(it).remote_ack_[ hdr->msg_nb(i).id() ] = hdr->msg_nb(i).delivery();

		// Computes metric of straight link between current node and its neighbor.
		if ( hdr->msg_nb(i).id() == id() ) {

			if ( nb(it).local_ack().ratio() != 0 && hdr->msg_nb(i).delivery() != 0 ) {
				alt = float ( 1.0 / ( nb(it).local_ack().ratio() * hdr->msg_nb(i).delivery() ));
			}
			else {
				alt = INF_METRIC;
			}

			if ( alt < best_metric ) {
				best_metric = alt;
				best_hop	= hdr->pkt_src();
			}
		}

		// Computes metric of path to neighbor using another neighbor as relay.
		else {

			itn = nb().find( hdr->msg_nb(i).id() );

			if ( itn != nb().end() ) {

				ita = nb(itn).remote_ack().find( hdr->pkt_src() );

				if (ita != nb(itn).remote_ack().end() ) {

					if ( nb(itn).remote_ack(ita) != 0 &&
						hdr->msg_nb(i).delivery() != 0 &&
						nb(itn).next_hop() != hdr->pkt_src() ) {

						alt = float (( 1.0 / (nb(itn).remote_ack(ita) * hdr->msg_nb(i).delivery())) + nb(itn).metric());
					}
					else {
						alt = INF_METRIC;
					}
					if ( alt < best_metric ) {
						best_metric = alt;
						best_hop	= nb(itn).next_hop();
					}
				}
			}
		}
	}

	nb(it).metric()		= best_metric;
	nb(it).next_hop()	= best_hop;

	return best_metric;
}


///
/// \brief Updates gateways states.
///
/// \param hdr content of a LIBR message received from a neighbor
/// \param hop_metric next hop metric of the announced distance vector route.
///
void
LIBR_state::updateGateways(hdr_LIBR_pkt* hdr , float hop_metric) {

	LIBR_gw_state*	temp;

	// Verification of valid gateway and one-hop loop.
	if (hdr->msg_pri_gw().id() != NULL_ID	&&
		hdr->msg_pri_gw().id() != id()		&&
		hdr->msg_pri_gw().next_hop() != id()) {

		// Checks if it is a known gateway.
		if ( hdr->msg_pri_gw().id() == pri_gw()->id() ) {
			writeData( hdr->msg_pri_gw() , pri_gw() , hop_metric , hdr->pkt_src() );
		}
		else if ( hdr->msg_pri_gw().id() == sec_gw()->id() ) {
			writeData( hdr->msg_pri_gw() , sec_gw() , hop_metric , hdr->pkt_src() );
		}
		// Saves announced gateway if it has a metric better than the current one.
		else if (hdr->msg_pri_gw().metric() != INF_METRIC &&
				 hop_metric != INF_METRIC &&
				(hdr->msg_pri_gw().metric() + hop_metric) < pri_gw()->metric() ) {
			pri_gw()->id() = hdr->msg_pri_gw().id();
			writeData( hdr->msg_pri_gw() , pri_gw() , hop_metric , hdr->pkt_src() );

			// Flag to announce new primary gateway.
           	new_pri_gw() = TRUE;

		}
		else if ( (hdr->msg_pri_gw().metric() + hop_metric) < sec_gw()->metric()) {
			sec_gw()->id() = hdr->msg_pri_gw().id();
			writeData( hdr->msg_pri_gw() , sec_gw() , hop_metric , hdr->pkt_src() );
		}
	}

	// Verification of valid gateway and one-hop loop.
	if (hdr->msg_sec_gw().id() != NULL_ID	&&
		hdr->msg_sec_gw().id() != id()		&&
		hdr->msg_sec_gw().next_hop() != id()) {
		
		// Checks if it is a known gateway.
		if ( hdr->msg_sec_gw().id() == pri_gw()->id() ) {
			writeData( hdr->msg_sec_gw() , pri_gw() , hop_metric , hdr->pkt_src() );
		}
		else if ( hdr->msg_sec_gw().id() == sec_gw_->id() ) {
			writeData( hdr->msg_sec_gw() , sec_gw() , hop_metric , hdr->pkt_src() );
		}
		// Saves announced gateway if it has a metric better than the current one.
		else if (hdr->msg_sec_gw().metric() != INF_METRIC &&
				 hop_metric != INF_METRIC &&
				(hdr->msg_sec_gw().metric() + hop_metric) < sec_gw()->metric() ) {
			sec_gw()->id() = hdr->msg_sec_gw().id();
			writeData( hdr->msg_sec_gw() , sec_gw() , hop_metric , hdr->pkt_src() );
		}
	}

	// Reorders primary and secondary gateways according to the best metric.
	if ( sec_gw()->metric() < pri_gw()->metric() ) {
		temp = pri_gw();
		pri_gw() = sec_gw();
		sec_gw() = temp;
	}
	
}

///
/// \brief Updates gateways's metric and next hop.
///
/// \param msg content related to gateway of a LIBR message
/// \param gw gateway state
/// \param hop_metric next hop metric of the announced distance vector route.
/// \param pkt_src id of neighbor that sent LIBR message
///
void
LIBR_state::writeData( LIBR_msg_gw msg, LIBR_gw_state* gw , float hop_metric , libr_id_t pkt_src ) {

	// First update
	if ( gw->metric() == INF_METRIC ) {

		gw->metric()		= msg.metric() + hop_metric;
		gw->next_hop()      = pkt_src;
		gw->seq()           = msg.seq();
		gw->timeouts() 		= 0;
		gw->hop_timeouts()	= 0;
		gw->alt_metric()	= INF_METRIC;
		gw->alt_hop()		= NULL_ID;
	}
	
	// Update with new a sequence number.
	else if ( newSeqNum(msg.seq(),gw->seq()) ) {

		// Changes next hop in case of inactive node or better alternative route.
		if (gw->hop_timeouts() >= LIBR_INACTIVE_HOP	||
			gw->alt_metric() < gw->metric() ) {

			gw->metric()		= gw->alt_metric();
			gw->next_hop()		= gw->alt_hop();
			gw->hop_timeouts()	= 0;
		}
		// Updates immediatly the route if it is from current next hop.
		if ( pkt_src == gw->next_hop() ) {

			gw->metric()		= msg.metric() + hop_metric;
			gw->hop_timeouts()	= 0;
		}

		// Update gateway sequence number.
		gw->seq()           = msg.seq();
		
		// Saves alternative route.
		gw->alt_metric()	= msg.metric() + hop_metric;
		gw->alt_hop()		= pkt_src;
		
		// Updates gateway and next hop timeouts.
		gw->timeouts() 		= 0;
		gw->hop_timeouts()	++;
	}
	
	// Update with the same sequence number.
	else if ( msg.seq() == gw->seq() ) {

		// Updates immediatly the route if it is from current next hop.
		if ( pkt_src == gw->next_hop() ) {

			gw->metric()		= msg.metric() + hop_metric;
			gw->hop_timeouts()	= 0;
		}
		// Checks if it is a one-hop loop.
		else if ( msg.next_hop() != id() ) {
			
			// Saves the best alternative hop.
			if (( msg.metric() + hop_metric ) < gw->alt_metric() ) {

				gw->alt_metric()	= msg.metric() + hop_metric;
				gw->alt_hop()       = pkt_src;
			}
			// Prefers the next hop that is closer to destination if it has 
			// the same metric of the alternative route.
			else if (( msg.metric() + hop_metric ) == gw->alt_metric() &&
					   abs(pkt_src - gw->id()) < abs(gw->alt_hop() - gw->id()) ) {

				gw->alt_metric()	= msg.metric() + hop_metric;
				gw->alt_hop()       = pkt_src;
			}
		}
	}
}


///
/// \brief Updates routes of neighbor and gateways into routing table.
///
/// \param it map iterator with position of related neighbor.
///
void
LIBR_state::updateRoutes(LIBR_nb_it it) {

	if ( nb(it).metric() != INF_METRIC )
		rtable().add_entry( ID2IP(nb_id(it)) , ID2IP(nb(it).next_hop()) );
	else
		rtable().rm_entry( ID2IP(nb_id(it)) );

	if ( pri_gw()->id() != NULL_ID ) {
		if ( pri_gw()->metric() != INF_METRIC )
			rtable().add_entry( ID2IP(pri_gw()->id()) , ID2IP(pri_gw()->next_hop()) );
		else
			rtable().rm_entry( ID2IP(pri_gw()->id()) );
	}

	if ( sec_gw()->id() != NULL_ID ) {
		if ( sec_gw()->metric() != INF_METRIC )
			rtable().add_entry( ID2IP(sec_gw()->id()) , ID2IP(sec_gw()->next_hop()) );
		else
			rtable().rm_entry(ID2IP(sec_gw()->id()));
	}
}


///
/// \brief Checks if the received sequence number is a newer sequence number.
///
/// \param msg_seq sequence number of gateway in the LIBR message.
/// \param node_seq known sequence number of gateway.
///
/// \return TRUE or FALSE
///
bool
LIBR_state::newSeqNum( u_int8_t msg_seq , u_int8_t node_seq ) {

	if ( ( u_int8_t ( msg_seq - node_seq - 1 ) ) < LIBR_PKT_WINDOW	)
		return TRUE;
	else
		return FALSE;

}


///
/// \brief Returns the IP address of the closest neighbor to destination.
///
/// \param dest IP address of destination.
///
/// \return IP address of the closest neighbor to destination.
///
nsaddr_t
LIBR_state::closestID( nsaddr_t dest ) {

	LIBR_nb_it		first		= nb().begin();
	LIBR_nb_it		last		= --nb().end();
	libr_id_t 		dest_id		= IP2ID(dest);
	libr_id_t 		min_dist	= NULL_ID;
	libr_id_t 		closest_nb;

	
	if ( dest_id < nb_id(first) && nb(first).metric() != INF_METRIC) {
		return ID2IP(nb_id(first));
	}
	else if ( dest_id > nb_id(last) && nb(last).metric() != INF_METRIC) {
		return ID2IP(nb_id(last));
	}
	else {
		for ( LIBR_nb_it it = nb().begin(); it != nb().end(); ++it ) {
			if ( abs(nb_id(it) - dest_id) < min_dist && nb(it).metric() != INF_METRIC) {
				min_dist = abs(nb_id(it) - dest_id) ;
				closest_nb = nb_id(it);
			}
		}
		return ID2IP(closest_nb);
	}
}


///
/// \brief Prints out the content of the node state to a given trace file.
///
/// \param out the trace where the node state must be written into.
///
void
LIBR_state::print(Trace* out) {
	
	sprintf(out->pt_->buffer(), 
		"\tNode ID: %d; GW_pri[ID: %d, next_hop: %d, metric: %f, seq_num: %d]; GW_sec[ID: %d, next_hop: %d, metric: %f, seq_num: %d] ",
		id(), pri_gw()->id(), pri_gw()->next_hop(), pri_gw()->metric(), pri_gw()->seq(),
		sec_gw()->id(), sec_gw()->next_hop(), sec_gw()->metric(), sec_gw()->seq());
	
	out->pt_->dump();

	for ( LIBR_nb_it it = nb().begin(); it != nb().end(); ++it ) {
		sprintf(out->pt_->buffer(), "\t\tNeighboor %d (next_hop: %d, metric: %f)",
			nb_id(it), nb(it).next_hop(), nb(it).metric());
		out->pt_->dump();
	}
		
}
