/****************************************************************************
 * Universidade Federal Fluminense - Instituto de Computao                *
 * Author: Bruno Siqueira                                                   *
 * brunos@midiacom.uff.br                                                   *
 * 2013                                                                     *
 ****************************************************************************/

///
/// \file	LIBR_state.h
/// \brief	This header file declares and defines internal state of an LIBR node.
///

#ifndef __LIBR_state_h__
#define __LIBR_state_h__


#include <limits>
#include <timer-handler.h>
#include "LIBR_rtable.h"
#include "LIBR_repositories.h"

/// Wildcard mask used for convertion from IP to ID.
#define WILDCARD_MASK		0x000000FF

/// IP network used for convertion from ID to IP.
#define IP_NETWORK 			0x00000000

/// Converts IP to ID.
#define IP2ID(ip_a)			libr_id_t (ip_a & WILDCARD_MASK)

/// Converts ID to IP.
#define ID2IP(id_a)			nsaddr_t  (id_a | IP_NETWORK   )

/// Reserved number of null ID.
#define NULL_ID				255

/// Reserved number of infinity metric (it is printed out as "inf").
#define INF_METRIC			std::numeric_limits<float>::infinity()

/// Number of timeouts to consider that a node is inactive.
#define LIBR_INACTIVE_NODE  4

/// Number of timeouts to delete the state of an inactive node.
#define LIBR_DELETED_NODE	8

/// Number of timeouts necessary to change the next hop of a destination.
#define LIBR_INACTIVE_HOP  	2


class LIBR;

/// Encapsulates data structures needed for maintaining internal state of a LIBR node.
class LIBR_state {

	friend class LIBR;
	friend class LIBR_UpdateTimer;

	libr_id_t		id_;			///< ID of agent
	LIBR_rtable		rtable_;		///< Routing table
	LIBR_nb_state	nb_;			///< Set of neighboors
	LIBR_gw_state*	pri_gw_;		///< Gateway with lowest metric
	LIBR_gw_state*	sec_gw_;		///< Gateway with second-lowest metric
	bool			new_pri_gw_;	///< Flag of new primary gateway

	inline libr_id_t&		id()				{ return id_; }
	inline LIBR_rtable&		rtable()			{ return rtable_; }
	inline LIBR_nb_state&	nb()				{ return nb_; }
	inline LIBR_gw_state*&	pri_gw()			{ return pri_gw_; }
	inline LIBR_gw_state*&	sec_gw()			{ return sec_gw_; }
	inline bool&			new_pri_gw()		{ return new_pri_gw_; }
	inline LIBR_nb_info&	nb(LIBR_nb_it i)	{ return i->second; }
	inline libr_id_t		nb_id(LIBR_nb_it i)	{ return i->first; }
		
	float		updateNeighboor(hdr_LIBR_pkt* , LIBR_nb_it);
	void		updateGateways(hdr_LIBR_pkt* , float);
	void		updateRoutes(LIBR_nb_it);
	void		writeData(LIBR_msg_gw , LIBR_gw_state* , float , libr_id_t);
	bool		newSeqNum(u_int8_t, u_int8_t);

protected:
	void		update(hdr_LIBR_pkt*);
	void		checkup();
	nsaddr_t	closestID(nsaddr_t);

public:
	LIBR_state(libr_id_t);
	void		print(Trace*);
	
};

#endif  /* __LIBR_state_h__ */
